/*
 * executer.cpp
 *
 *      Author: David Fernandez Lopez
 *      Debris Collector Core
 */

#include <vector>
#include <cmath>
#include "model.h"
#include "elastic.h"
#include "ros/ros.h"
#include "debris_collector/SpaceObjectMsg.h"
#include "debris_collector/PlanTaskMsg.h"
#include "debris_collector/MovementMsg.h"
#include "debris_collector/LocationMsg.h"
#include "debris_collector/GrabObjectSrv.h"
#include "std_srvs/Empty.h"

/*State Definitions*/
#define IDLE_STATE 0
#define SELECT_NEXT 1
#define MOVE_TO_NEXT 2
#define GRAB 3
#define MOVE_TO_DROP 4
 
//Parameters
#define DROP_SPEED 45.0f
#define REACH 250.0f
#define SAFETY_DISTANCE 100.0f
#define MAX_SPEED 50.0f
#define WAIT_TIME 15

std::vector<ProjectableSpaceObject*> objects;
std::vector<Task*> tasks;
Robot* robot = NULL;
int state;
bool data;

//Mesage callback
void locationCallback(const debris_collector::LocationMsg& msg){
	*robot->current_point->r = msg.r;
	*robot->current_point->polar = msg.polar;
	*robot->current_point->azimuth = msg.azimuth;
	data = true;
}

void objectsCallback(const debris_collector::SpaceObjectMsg& msg){
	//ROS_INFO_STREAM("Object Message recieved");
	for(unsigned int i=0; i<objects.size(); i++){
		if(*objects[i]->id == msg.id){
			//If the object is already in the targets vector some parameters should be updated
			objects[i]->orbit->updateStartingPoint(msg.r,msg.polar,msg.azimuth);
			*objects[i]->current_point->r = msg.r;
			*objects[i]->current_point->polar = msg.polar;
			*objects[i]->current_point->azimuth = msg.azimuth;
			objects[i]->active = true;
			return;
		}
	}
	//If the object has not yet appeared in the targets vector it has to be inserted
	if(msg.id%2!=0){
		objects.push_back(new ProjectableTargetSO(msg.id,msg.r,msg.polar,msg.azimuth,msg.v,msg.alpha,msg.beta,msg.mass));
	}else{
		objects.push_back(new ProjectableObstacleSO(msg.id,msg.r,msg.polar,msg.azimuth,msg.v,msg.alpha,msg.beta));
	}
}

void planCallback(const debris_collector::PlanTaskMsg& msg){
	if(msg.task_number==0){
		//Clean all tasks
		while(tasks.size()>0)
			tasks.erase(tasks.begin());
	}
	tasks.push_back(new Task(msg.target_id, msg.destination_time, objects));
	if(msg.total_tasks == tasks.size())
		state = SELECT_NEXT;
}

/*General use methods*/
void compute_force(ProjectableSpaceObject* so, Point* robot, EuclideanVector* vr, EuclideanVector* resulting_force){
	//Calculate closeness vecotr
	EuclideanVector* D = new EuclideanVector();
	distance_vector(so->current_point, robot, D);
	closeness(D);

	//Calculate the object's speed vector
	EuclideanVector* vo = new EuclideanVector();
	orbit_speed(so, vo);

	//Calculate the cross_product
	EuclideanVector* f0 = new EuclideanVector();
	EuclideanVector* f1 = new EuclideanVector();
	cross_product(D, vo, f0, f1);

	//The least disruptive is chosen
	if(choose(f0,f1,vr)==0){
		//The least disruptive is f0
		resulting_force->copy_data(f0);
		//cout << "F0 chosen" << endl;
	}else{
		//The least disruptive is f1
		resulting_force->copy_data(f1);
		//cout << "F1 chosen" << endl;
	}

	delete D;
	delete vo;
	delete f0;
	delete f1;
}

bool move_to_target(Task* task, ros::Publisher mov_pub, float t){
	bool result = true;
	//Calculate the movement vector to the target
	EuclideanVector* vr = new EuclideanVector();
	EuclideanVector* original = new EuclideanVector();
	distance_vector(robot->current_point, task->target_point, vr);
	float time_interval = *task->time-t;
	if(time_interval <0.0f){
		//The robot is late
		time_interval = 1.0f;
	}
	*vr->x = *vr->x / (time_interval);
	*vr->y = *vr->y / (time_interval);
	*vr->z = *vr->z / (time_interval);
	original->copy_data(vr);

	//For each object in close proximity calculate its repulsive force and modify the movement vector accordingly
	for(unsigned int i=0; i< objects.size(); i++){
		if(distance(objects[i]->current_point, robot->current_point)<SAFETY_DISTANCE && *objects[i]->id != *task->target_id){
			EuclideanVector* repulsive_force = new EuclideanVector();
			compute_force(objects[i], robot->current_point, original, repulsive_force);
			*vr->x += *repulsive_force->x;
			*vr->y += *repulsive_force->y;
			*vr->z += *repulsive_force->z;
			delete repulsive_force;
		}
	}

	float speed = norm(vr);
	debris_collector::MovementMsg msg;
	*vr->x = *vr->x/speed;
	*vr->y = *vr->y/speed;
	*vr->z = *vr->z/speed;
	if(speed <= MAX_SPEED){
		msg.speed = speed;
	}else{
		msg.speed = MAX_SPEED;
		result = false;
	}
	quaternion(robot->movement, vr, &msg.x, &msg.y, &msg.z, &msg.w);
	mov_pub.publish(msg);
	robot->movement->copy_data(vr);
	ROS_INFO_STREAM("Movement message published ("<<*vr->x<<","<<*vr->y<<","<<*vr->z<<") "<<speed);

	delete vr;
	delete original;

	//If speed is smaller or equal to MAX_SPEED return true
	return result;
}

int main(int argc, char **argv){
	//Loops per second, Hertz
	int time_ratio = 1;

	state = IDLE_STATE;
	Task* current_task;
	Task* drop = new Task(-1, 0.0f, objects);
	delete drop->target_point;
	drop->target_point = new Point(200000.0f,45.0f,45.0f);
	robot = new Robot(100000.0f,0.0f,0.0f,1.0f/std::sqrt(3.0f),1.0f/std::sqrt(3.0f),1.0f/std::sqrt(3.0f));
	data = false;

	//ROS init
	ros::init(argc, argv, "ExecuterNode");
	ros::NodeHandle n;

	//ROS communication init
	ros::Subscriber subObj = n.subscribe("Objects", 1000, objectsCallback);
	ros::Subscriber subPla = n.subscribe("Plan", 1000, planCallback);
	ros::Subscriber subLoc = n.subscribe("Location", 1000, locationCallback);
	ros::Publisher mov_pub = n.advertise<debris_collector::MovementMsg>("Move", 1000);
	ros::ServiceClient client_grab = n.serviceClient<debris_collector::GrabObjectSrv>("Grab");
	ros::ServiceClient client_plan = n.serviceClient<std_srvs::Empty>("NewTaskList");
	std_srvs::Empty srv_plan;

	//current time
	float t = 0.0f;
	//Counter for time between NewPlanList calls
	int time_calls = -1;

	//ROS loop rate sets sleep time
	ros::Rate loop_rate(time_ratio);

	//The node loop, it will go on until the node is killed .
	//ros::ok() returns false if: a SIGINT is received (Ctrl-C), The node has been kicked off the network by another node with the same name,
	//ros::shutdown() has been called by another part of the application or all ros::NodeHandles have been destroyed.
	while(ros::ok()){
		if(data){
			//Choose action according to current state
			switch(state){
				case IDLE_STATE:
				{
					ROS_INFO_STREAM("IDLE_STATE");	
					//Call the planner node to provide a new plan every few seconds				
					if((time_calls<0 || time_calls>WAIT_TIME)){
						if(!client_plan.call(srv_plan))
							ROS_ERROR_STREAM("Failed to call New Plan service");
						else
							time_calls = 0;
					}
					time_calls++;					
				}
					break;
				case SELECT_NEXT:
				{
					ROS_INFO_STREAM("SELECT_NEXT");
					//Prepare the next task to be executed
					if(tasks.size()>0){
						current_task = tasks[0];
						state = MOVE_TO_NEXT;
						t = 0.0f;
						for(unsigned int i=0; i<objects.size(); i++){
							if(*objects[i]->id == *current_task->target_id){
								*current_task->target_point->r = *objects[i]->project(*current_task->time)->r;
								*current_task->target_point->polar = *objects[i]->project(*current_task->time)->polar;
								*current_task->target_point->azimuth = *objects[i]->project(*current_task->time)->azimuth;
								break;
							}
						}
						move_to_target(current_task, mov_pub, t);
					}else{
						//There are no tasks on the list, this shouldn't be the current state
						ROS_ERROR_STREAM("No tasks to be done");
						state = IDLE_STATE;
					}
				}
					break;
				case MOVE_TO_NEXT:
				{
					ROS_INFO_STREAM("MOVE_TO_NEXT - ("<<*current_task->target_point->r<<","<<*current_task->target_point->polar<<","<<*current_task->target_point->azimuth<<") r("<<*robot->current_point->r<<","<<*robot->current_point->polar<<","<<*robot->current_point->azimuth<<")");
					//Move towards the target point
					if(!move_to_target(current_task, mov_pub, t)){
						//If the required speed is to high move to the drop point
						state = MOVE_TO_DROP;
						current_task = drop;
					}
					//If the robot is in close proximity to the target grab it
					if(distance(robot->current_point, current_task->target_point)<REACH){
						state = GRAB;
						t = 0.0f;
					}
				}
					break;
				case GRAB:
				{
					ROS_INFO_STREAM("GRAB");
					//Try to grab the target
					debris_collector::GrabObjectSrv srv_grab;
					srv_grab.request.id = *current_task->target_id;
					if(client_grab.call(srv_grab)){
						if(srv_grab.response.result == 0){
							//Target succesfully captured
							tasks.erase(tasks.begin());
							if(tasks.size()!=0){
								//More targets to pick up
								state = SELECT_NEXT;
							}else{
								//All targets captured, move to drop point
								current_task = drop;
								state = MOVE_TO_DROP;
								/*Calculate time to drop point*/
								*current_task->time = distance(robot->current_point, current_task->target_point)/DROP_SPEED;
								t = 0.0f;
							}
						}else{
							//Grab not succesfull, try to get closer to the target current position
							move_to_target(current_task, mov_pub, t);
						}
					}else{
						//The service call has failed
						ROS_ERROR_STREAM("Failed to call Grab service");
					}
				}
					break;
				case MOVE_TO_DROP:
				{
					ROS_INFO_STREAM("MOVE_TO_DROP");
					//Move to the drop point
					move_to_target(current_task, mov_pub, t);
					//If the drop point is in close proximity, stop
					if(distance(robot->current_point, current_task->target_point)<REACH){
						state = IDLE_STATE;
					}
				}
					break;
				default:
					//An unexpected state value has been received
					ROS_ERROR_STREAM("Incorrect state: "<< state);
					state = IDLE_STATE;
			}

			//Clean undected objects
			for(unsigned int i=0; i< objects.size(); i++){
				if(objects[i]->active){
					//No cleanup needed
					objects[i]->active = false;
				}else{
					//Cleanup needed, the projectable object is removed from the vector
					objects.erase(objects.begin()+i);
				}
			}

			//Update timer
			t = t +1.0f;
		}else{
			ROS_INFO_STREAM("No data recieved yet.");
		}

		//All pending messages are read and its callback methods called
		ros::spinOnce();
		loop_rate.sleep();
	}

	delete drop;
	delete robot;
	for(int i=0; i<objects.size(); i++)
		delete objects[i];
	objects.clear();
	std::vector<ProjectableSpaceObject*>().swap(objects);
	for(int i=0; i<tasks.size(); i++)
		delete tasks[i];
	tasks.clear();
	std::vector<Task*>().swap(tasks);

	return 0;
}



