/*
 * genetics.cpp
 *
 *      Author: David Fernandez Lopez
 *      Debris Collector Core
 */

#include <vector>
#include <cstdlib>
#include <cmath>
#include "genetics.h"
#include "model.h"

#define MAX_TIME_STEP 86340
#define MIN_TIME_STEP 3600.0f
#define MIN_GRAD 0.005f
#define MAX_CARRIED_OBJ 5
#define PI 3.14159265f

using namespace std;

/*******************************************/
/************ General use methods **********/
/*******************************************/

float random_float(float a, float b) {
    float random = ((float) rand()) / (float) RAND_MAX;
    float diff = b - a;
    float r = random * diff;
    return a + r;
}

float loadTime(Point* a, ProjectableTargetSO* b, float t){
	float time_point = t + MIN_TIME_STEP;
	float desc_step = 10.0f;
	float asc_step  = 20.0f;
	float x = *a->r * sin(*a->polar*PI/180.0f) * cos(*a->azimuth*PI/180.0f);
	float y = *a->r * sin(*a->polar*PI/180.0f) * sin(*a->azimuth*PI/180.0f);
	float z = *a->r * cos(*a->polar*PI/180.0f);
	bool grad_desc = false;
	bool found = false;
	float last_grad = 0.0f;

	while(!found){
		float sp = sin((*b->orbit->starting_point->polar + *b->orbit->alpha*time_point)*PI/180.0f);
		float cp = cos((*b->orbit->starting_point->polar + *b->orbit->alpha*time_point)*PI/180.0f);
		float sa = sin((*b->orbit->starting_point->azimuth + *b->orbit->beta*time_point)*PI/180.0f);
		float ca = cos((*b->orbit->starting_point->azimuth + *b->orbit->beta*time_point)*PI/180.0f);
		float grad = -(((*b->project(time_point)->r)*cp - z)*(*b->orbit->alpha)*(*b->project(time_point)->r)*sp
				- ((*b->project(time_point)->r)*sp*ca - x)*((*b->orbit->alpha)*(*b->project(time_point)->r)*ca*cp - (*b->orbit->beta)*(*b->project(time_point)->r)*sa*sp)
				- ((*b->project(time_point)->r)*sa*sp -y)*((*b->orbit->alpha)*(*b->project(time_point)->r)*sa*cp + (*b->orbit->beta)*(*b->project(time_point)->r)*sp*ca)
			)/sqrt(pow(((*b->project(time_point)->r)*cp - z),2.0f) + pow(((*b->project(time_point)->r)*sp*ca - x),2.0f) + pow(((*b->project(time_point)->r)*sa*sp - y),2.0f));
		if(!grad_desc && grad > 0){
			//cost is increasing, a new point has to be selected to start the algorithm
			//As the cost is increasing a gradient ascent will be performed, trying to overshoot the maximum to start the gradient descent
			time_point += asc_step * grad;
		}else{
			//cost is decreasing, so the algorithm is applied
			grad_desc = true;
			if(grad>0 && last_grad<0){
				//The minimum point is between those two values, so the step_size should be reduced
				desc_step /= 2;
			}
			if(abs(grad)<MIN_GRAD)
				found = true;
			else
				time_point -= desc_step * grad;
			last_grad = grad;
		}
	}

	return time_point;
}


/*******************************************/
/****************** Classes ****************/
/*******************************************/

/**Class roulette**/
Roulette::Roulette(int max){
	ball = new Number;
	ball->number = 0;
	Number* tmp = ball;
	for(int i=1; i<max; i++){
		ball->next = new Number;
		ball = ball->next;
		ball->number = i;
	}
	ball->next = tmp;
}

Roulette::~Roulette(){
	while(ball!=NULL)
		roll(0);
}

int Roulette::roll(int passes){
	int result = -1;
	if(ball!=NULL){
		if(ball != ball->next){
			for(int i=0; i<passes; i++)
				ball = ball->next;
			result = ball->next->number;
			Number* tmp = ball->next;
			ball->next = ball->next->next;
			delete tmp;
		}else{
			result = ball->number;
			delete ball;
			ball = NULL;
		}
	}
	return result;
}

/*Class Individual*/
Individual::Individual(int size){
	total_cost = new float;
	*total_cost = 0.0f;
	fitness = new float;
	*fitness = 0.0f;
	order.resize(min(size,MAX_CARRIED_OBJ));
	time.resize(min(size,MAX_CARRIED_OBJ));
	cost.resize(min(size,MAX_CARRIED_OBJ));
}

Individual::Individual(int size, const vector<ProjectableTargetSO*> &points, Point* robot){
	total_cost = new float;
	*total_cost = 0.0f;
	fitness = new float;
	*fitness = 0.0f;
	order.resize(min(size,MAX_CARRIED_OBJ));
	time.resize(min(size,MAX_CARRIED_OBJ));
	cost.resize(min(size,MAX_CARRIED_OBJ));
	Roulette* roulette = new Roulette(size);
	for(int j=0; j<min(size,MAX_CARRIED_OBJ); j++){
		order[j] = roulette->roll(rand()%order.size());
	}
	delete roulette;
	//Load times
	time[0] = loadTime(robot, points[order[0]], 0.0f);
	for(unsigned int j=1; j<order.size(); j++)
		time[j] = loadTime(points[order[j-1]]->project(time[j-1]), points[order[j]], time[j-1]);
}

Individual::~Individual(){
	delete total_cost;
	delete fitness;
	order.clear();
	std::vector<int>().swap(order);
	time.clear();
	std::vector<float>().swap(time);
	cost.clear();
	std::vector<float>().swap(cost);
}

float Individual::cost_to_next(int current){
	for(unsigned int i=0; i<order.size()-1; i++){
		if(order[i]==current)
			return cost[i+1];
	}
	//If this point is reached then current is the last one in order, and has no associated cost
	return 0.0f;
}

int Individual::next(int current){
	for(unsigned int i=0; i<order.size()-1; i++){
		if(order[i]==current)
			return order[i+1];
	}
	//If this point is reached then current is the last one in order, so next is initial (cyclic)
	return 0;
}

float Individual::next_time(int current){
	for(unsigned int i=0; i<time.size()-1; i++){
		if(time[i]==current)
			return time[i+1]-time[i];
	}
	//If this point is reached then current is the last one in order, so next is initial (cyclic)
	return time[0];
}

void Individual::compute_costs(const vector<ProjectableTargetSO*> &points, Point* robot){
	cost[0] = distance(robot, points[order[0]]->project(time[0]))/time[0];
	*total_cost = cost[0];
	for(unsigned int i=1; i<order.size(); i++){
		cost[i] = distance(points[order[i-1]]->project(time[i-1]),points[order[i]]->project(time[i]))/(time[i]-time[i-1]);
		*total_cost += cost[i];
	}
	*fitness = 1.0f/(*total_cost);
}

bool Individual::loadValue(int elem, float new_time, int pos){
	for(int i=0; i<pos; i++){
		if(order[i]==elem) return false;
	}
	//If this point is reached the elem is not present in the array
	order[pos] = elem;
	time[pos] = time[pos-1]+new_time;
	return true;
}

bool Individual::loadValue(int elem, int pos){
	for(int i=0; i<pos; i++){
		if(order[i]==elem) return false;
	}
	//If this point is reached then elem is not present in the array
	order[pos] = elem;
	return true;
}

void Individual::loadRandomValue(int pos){
	Roulette* r = new Roulette(order.size());
	int next;
	do{
		next = r->roll(rand()%order.size());
	}while(!loadValue(next, pos));
	time[pos] = MIN_TIME_STEP + ((pos==0)?((float)(rand()%MAX_TIME_STEP)):(time[pos-1]+((float)(rand()%MAX_TIME_STEP))));
	delete r;
}

void Individual::mutation(int pos1, int pos2){
	int temp = order[pos1];
	order[pos1] = order[pos2];
	order[pos2] = temp;
}

void Individual::migration(const vector<ProjectableTargetSO*> &points, Point* robot){
	Roulette* roulette = new Roulette(points.size());
	for(int j=0; j<order.size(); j++){
		order[j] = roulette->roll(rand()%order.size());
	}
	delete roulette;
	//Load times
	time[0] = loadTime(robot, points[order[0]], 0.0f);
	for(unsigned int j=1; j<order.size(); j++)
		time[j] = loadTime(points[order[j-1]]->project(time[j-1]), points[order[j]], time[j-1]);
}

/*Class Population*/
Population::Population(int size, const vector<ProjectableTargetSO*> &points, Point* robot){
	best_cost = new float;
	*best_cost = 10000000000000.0f;
	for(int i=0; i<size*(size-1); i++){
		individuals.push_back(new Individual(size, points, robot));
	}
}

Population::Population(int size){
	best_cost = new float;
	*best_cost = 10000000000000.0f;
	for(int i=0; i<size*(size-1); i++){
		individuals.push_back(new Individual(size));
	}
}

Population::~Population(){
	for(int i=0; i<individuals.size(); i++)
		delete individuals[i];
	individuals.clear();
	std::vector<Individual*>().swap(individuals);
	delete best_cost;
}

int Population::chooseParent(){
	//First a random float is generated between 0.0 and 1.0
	float chosen = (float)rand()/(float)RAND_MAX;

	//Then, the first individual with higher fitness is chosen
	for(unsigned int i=0; i<individuals.size(); i++){
		if(chosen < *individuals[i]->fitness)
			return i;
	}
	//If this point is reached an error has occurred
	return -1;
}

int Population::compute_costs(const vector<ProjectableTargetSO*> &points, Point* robot){
	float fitness_sum = 0.0f;
	int best = 0;
	for(unsigned int i=0; i<individuals.size(); i++){
		individuals[i]->compute_costs(points, robot);
		fitness_sum += *individuals[i]->fitness;
		if(*best_cost > *individuals[i]->total_cost){
			*best_cost = *individuals[i]->total_cost;
			best = i;
		}
	}
	
	//Now a fitness factor is calculated for each individual.
	*individuals[0]->fitness = *individuals[0]->fitness/fitness_sum;
	for(unsigned int j=1; j<individuals.size(); j++){
		*individuals[j]->fitness = *individuals[j-1]->fitness + (*individuals[j]->fitness/fitness_sum);
	}

	return best;
}

