/*
 * locator.cpp
 *
 *      Author: David Fernandez Lopez
 *      Debris Collector Core
 */

#include <cstdlib>
#include <ctime>
#include <vector>
#include "ros/ros.h"
#include "model.h"
#include "particle.h"
#include "debris_collector/BeaconMsg.h"
#include "debris_collector/MovementMsg.h"
#include "debris_collector/LocationMsg.h"

//Parameters
#define MAX_ERROR 1000.0f

float speed = 0.5f;
ParticleList* particle_list = NULL;
bool data;
std::vector<std::vector<float> > rotation;

/*Callback methods*/
void moveCallback(const debris_collector::MovementMsg& msg){
	quaternion_to_rot(msg.x, msg.y, msg.z, msg.w, rotation);
	speed = msg.speed;
}

void beaconCallback(const debris_collector::BeaconMsg& msg){
	data = true;
	if(particle_list!=NULL){
		for(unsigned int j=0; j<particle_list->particles.size(); j++){
			particle_list->particles[j]->loadDistance(msg.beacon_r, msg.beacon_polar, msg.beacon_azimuth, msg.distance);
		}
	}
		
}

int main(int argc, char **argv){
	//Loops per second, Hertz
	int time_ratio = 1;
	data = false;

	std::srand(time(0));

	//ROS init
	ros::init(argc, argv, "LocationNode");
	ros::NodeHandle n;

	//ROS communication init
	ros::Publisher loc_pub = n.advertise<debris_collector::LocationMsg>("Location", 1000);
	ros::Subscriber subMov = n.subscribe("Move", 1000, moveCallback);
	ros::Subscriber subBea = n.subscribe("Beacon", 1000, beaconCallback);

	/*Build the initial particle vector*/
	particle_list = new ParticleList(true);
	Robot* com_robot = new Robot();
	bool located = false;

	//ROS loop rate sets sleep time
	ros::Rate loop_rate(time_ratio);

	rotation.resize(3);
	rotation[0].resize(3);
	rotation[1].resize(3);
	rotation[2].resize(3);


	//The node loop, it will go on until the node is killed .
	//ros::ok() returns false if: a SIGINT is received (Ctrl-C), The node has been kicked off the network by another node with the same name,
	//ros::shutdown() has been called by another part of the application or all ros::NodeHandles have been destroyed.
	while(ros::ok()){
		if(data){
			double error = particle_list->move(speed, rotation);

			/*Generate a new particle population*/
			ParticleList* new_particle_list = new ParticleList(false);
			for(unsigned int i=0; i<particle_list->particles.size(); i++){
				new_particle_list->particles[i]->load(particle_list->particles[particle_list->resample()]);
			}
			delete particle_list;
			particle_list = new_particle_list;

			//Decide if location should be
			if(error < MAX_ERROR || located){
				//Calculate estimated position and publish it
				particle_list->computeMedium(com_robot->current_point, com_robot->movement);
				debris_collector::LocationMsg msg;
				msg.r = *com_robot->current_point->r;
				msg.polar = *com_robot->current_point->polar;
				msg.azimuth = *com_robot->current_point->azimuth;
				msg.heading_x = *com_robot->movement->x;
				msg.heading_y = *com_robot->movement->y;
				msg.heading_z = *com_robot->movement->z;
				loc_pub.publish(msg);
				located = true;
				ROS_INFO_STREAM("Position ("<<msg.r<<","<<msg.polar<<","<<msg.azimuth<<") - ("<<msg.heading_x<<","<<msg.heading_y<<","<<msg.heading_z<<") e="<<error);
			}else{
				ROS_INFO_STREAM("Computing e="<<error);
			}
		}else{
			ROS_INFO_STREAM("No data recieved yet.");
		}

		//All pending messages are read and its callback methods called
		ros::spinOnce();
		loop_rate.sleep();
	}

	delete com_robot;
	delete particle_list;

	return 0;
}
