/*
 * model.cpp
 *
 *      Author: David Fernandez Lopez
 *      Debris Collector Core
 */

#include <cmath>
#include "model.h"

#define PI 3.14159265f

using namespace std;

/*******************************************/
/************ General use methods **********/
/*******************************************/

float distance(Point* a, Point* b){
    float xa= *a->r * sin(*a->polar*PI/180.0f) * cos(*a->azimuth*PI/180.0f);
    float ya= *a->r * sin(*a->polar*PI/180.0f) * sin(*a->azimuth*PI/180.0f);
    float za= *a->r * cos(*a->polar*PI/180.0f);
    float xb= *b->r * sin(*b->polar*PI/180.0f) * cos(*b->azimuth*PI/180.0f);
    float yb= *b->r * sin(*b->polar*PI/180.0f) * sin(*b->azimuth*PI/180.0f);
    float zb= *b->r * cos(*b->polar*PI/180.0f);
    return sqrt(pow(xa-xb,2)+pow(ya-yb,2)+pow(za-zb,2));
}

bool collision(Point* a, Point* b){
	if(*a->r == *b->r && *a->polar == *b->polar && *a->azimuth == *b->azimuth)
		return true;
	return false;
}

float to_deg(float angle){
	if(angle >= 360.0f){
		return angle - ((int)(angle/360)*360.0f);
	}else if(angle<0.0f){
		float rest = (angle + ((int)((-1.0f*angle)/360)*360.0f));
		if(rest==0.0f)
			return 0.0f;
		else
			return 360.0f + rest;
	}else
		return angle;
}

float norm(EuclideanVector* v){
	return sqrt(pow(*v->x, 2) + pow(*v->y, 2) + pow(*v->z, 2));
}

void quaternion_to_rot(float x, float y, float z, float w, vector<vector<float> >& rotation){
	//helper quantities - calculate these up front
	//to avoid redundancies
	float xSq = x * x;
	float ySq = y * y;
	float zSq = z * z;
	float wSq = w * w;
	float twoX = 2.0f * x;
	float twoY = 2.0f * y;
	float twoW = 2.0f * w;
	float xy = twoX * y;
	float xz = twoX * z;
	float yz = twoY * z;
	float wx = twoW * x;
	float wy = twoW * y;
	float wz = twoW * z;

	//fill in the first row
	rotation[0][0] = wSq + xSq - ySq - zSq;
	rotation[0][1] = xy - wz;
	rotation[0][2] = xz + wy;

	//fill in the second row
	rotation[1][0] = xy + wz;
	rotation[1][1] = wSq - xSq + ySq - zSq;
	rotation[1][2] = yz - wx;

    //fill in the third row
 	rotation[2][0] = xz - wy;
 	rotation[2][1] = yz + wx;
 	rotation[2][2] = wSq - xSq - ySq + zSq;
}

void apply_rotation(EuclideanVector* v1, std::vector<std::vector<float> >& rotation){
	float x = *v1->x;
	float y = *v1->y;
	float z = *v1->z;

	*v1->x = rotation[0][0]*x + rotation[0][1]*y + rotation[0][2]*z;
	*v1->y = rotation[1][0]*x + rotation[1][1]*y + rotation[1][2]*z;
	*v1->z = rotation[2][0]*x + rotation[2][1]*y + rotation[2][2]*z;
}


/*******************************************/
/****************** Classes ****************/
/*******************************************/

/**Class Point Methods**/

/*Class Point constructor. Both angles should be given in degrees.*/
Point::Point(float radial_distance, float polar_angle, float azimuthal_angle){
	r = new float;
	polar = new float;
	azimuth = new float;
	*r = radial_distance;
	*polar = to_deg(polar_angle);
	*azimuth = to_deg(azimuthal_angle);
}
Point::Point(){
	r = new float;
	polar = new float;
	azimuth = new float;
	*r = 0.0f;
	*polar = 0.0f;
	*azimuth = 0.0f;
}

Point::~Point(){
	delete r;
	delete polar;
	delete azimuth;
}

/**Class Orbit Methods**/

Orbit::Orbit(float r,float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed){
	v = new float;
	alpha = new float;
	beta = new float;
	starting_point = new Point(r,polar, azimuth);
	*v = speed;
	*alpha = polar_speed;
	*beta = azimuthal_speed;
}
Orbit::Orbit(){
	starting_point = new Point;
	v = new float;
	alpha = new float;
	beta = new float;
	*v = 0.0f;
	*alpha = 0.0f;
	*beta = 0.0f;
}

Orbit::~Orbit(){
	delete v;
	delete alpha;
	delete beta;
	delete starting_point;
}

Point Orbit::move(float* time){
	return Point(*starting_point->r+(*v**time),
			      *starting_point->polar+(*alpha**time),
			      *starting_point->azimuth+(*beta**time));
}

void Orbit::move(float* time, Point* p){
	*p->r = *starting_point->r+(*v**time);
	*p->polar = to_deg(*starting_point->polar+(*alpha**time));
    *p->azimuth = to_deg(*starting_point->azimuth+(*beta**time));
}

void Orbit::updateStartingPoint(float r, float polar, float azimuth){
	*starting_point->r = r;
	*starting_point->polar = polar;
	*starting_point->azimuth = azimuth;
}

/**Class SpaceObject Methods**/

SpaceObject::SpaceObject(int new_id, float r, float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed, float time){
	id = new int;
	starting_time = new float;
	current_point = new Point(r,polar, azimuth);
	orbit = new Orbit(r,polar, azimuth, speed, polar_speed, azimuthal_speed);
	*starting_time = time;
	*id = new_id;
}

SpaceObject::~SpaceObject(){
	delete starting_time;
	delete current_point;
	delete orbit;
	delete id;
}

/*Updates the object's position according to a general timeframe.
This method must recieve the current time since the model started,
not the elapsed time since the SpaceObject was created (this is 
calculated in this method)*/
void SpaceObject::updateSpaceObject(float* time){
	if(*time > *starting_time){
		float elapsed_time = *time-*starting_time;
		orbit->move(&elapsed_time, current_point);
	}
}

/*Test if the object exists at this time*/
bool SpaceObject::exist(float time){
	return time >= *starting_time;
}

/*Class TargetSO methods*/
TargetSO::TargetSO(int new_id, float r, float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed, float time, int mass):SpaceObject(new_id, r, polar, azimuth, speed, polar_speed, azimuthal_speed, time){
	object_mass = new int;
	*object_mass = mass;
}

TargetSO::~TargetSO(){
	delete object_mass;
}

/*Class ObstacleSO methods*/
ObstacleSO::ObstacleSO(int new_id, float r, float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed, float time):SpaceObject(new_id, r, polar, azimuth, speed, polar_speed, azimuthal_speed, time){
}

ObstacleSO::~ObstacleSO(){
}

/*class ProjectableSpaceObject methods*/
ProjectableSpaceObject::ProjectableSpaceObject(int new_id, float r, float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed):SpaceObject(new_id, r, polar, azimuth, speed, polar_speed, azimuthal_speed, 0.0f){
	projected_point = new Point();
	projected_time = new float;
	*projected_time = 0.0f;
	active = true;
}

ProjectableSpaceObject::~ProjectableSpaceObject(){
	delete projected_time;
	delete projected_point;
}

Point* ProjectableSpaceObject::project(float time){
	if(time!=*projected_time){
		*projected_time = time;
		float elapsed_time = time-*starting_time;
		orbit->move(&elapsed_time, projected_point);
	}
	return projected_point;
}

/*class ProjectableTargetSO methods*/
ProjectableTargetSO::ProjectableTargetSO(int new_id, float r, float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed, int mass):ProjectableSpaceObject(new_id, r, polar, azimuth, speed, polar_speed, azimuthal_speed){
	object_mass = new int;
	*object_mass = mass;
}

ProjectableTargetSO::~ProjectableTargetSO(){
	delete object_mass;
}

/*class ProjectableObstacleSO methods*/
ProjectableObstacleSO::ProjectableObstacleSO(int new_id, float r, float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed):ProjectableSpaceObject(new_id, r, polar, azimuth, speed, polar_speed, azimuthal_speed){
}

ProjectableObstacleSO::~ProjectableObstacleSO(){
}

/*class Task methods*/
Task::Task(int target, float dest_time, const vector<ProjectableSpaceObject*> &objects){
	target_id = new int;
	time = new float;
	*target_id = target;
	*time = dest_time;
	bool found = false;
	for(unsigned int i=0; i<objects.size() && !found; i++){
		if(*objects[i]->id == target){
			target_point = new Point(*objects[i]->project(*time)->r,*objects[i]->project(*time)->polar,*objects[i]->project(*time)->azimuth);
			found = true;
		}
	}
	if(!found)
		target_point = new Point();
}

Task::~Task(){
	delete target_id;
	delete time;
	delete target_point;
}

/*Class EuclideanVector methods*/
EuclideanVector::EuclideanVector(){
	x = new float;
	y = new float;
	z = new float;
	*x = 0.0f;
	*y = 0.0f;
	*z = 0.0f;
}
EuclideanVector::EuclideanVector(float x0, float y0, float z0){
	x = new float;
	y = new float;
	z = new float;
	*x = x0;
	*y = y0;
	*z = z0;
}
EuclideanVector::~EuclideanVector(){
	delete x;
	delete y;
	delete z;
}

void EuclideanVector::copy_data(EuclideanVector* v){
	*x = *v->x;
	*y = *v->y;
	*z = *v->z;
}

/*Class Robot methods*/
Robot::Robot(){
	current_point = new Point();
	movement = new EuclideanVector();
}

Robot::Robot(float r, float polar, float azimuth, float vx, float vy, float vz){
	current_point = new Point(r, polar, azimuth);
	movement = new EuclideanVector(vx, vy, vz);
}

Robot::~Robot(){
	delete current_point;
	delete movement;
}

void Robot::move(float speed){
	/*Translate to vector*/
	EuclideanVector* position = new EuclideanVector(*current_point->r * sin(*current_point->polar*PI/180.0f) * cos(*current_point->azimuth*PI/180.0f), *current_point->r * sin(*current_point->polar*PI/180.0f) * sin(*current_point->azimuth*PI/180.0f), *current_point->r * cos(*current_point->polar*PI/180.0f));

	/*Aply movement*/
	*position->x += speed*(*movement->x);
	*position->y += speed*(*movement->y);
	*position->z += speed*(*movement->z);

	/*Translate back*/
	*current_point->r = norm(position);
	*current_point->polar = acos(*position->z / *current_point->r) * 180.0f / PI;
	if(*position->x == 0.0f && *position->y == 0.0f)
		*current_point->azimuth = 0.0f;
	else if(*position->x >= 0.0f)
		*current_point->azimuth = asin(*position->y / sqrt(pow(*position->x,2.0f)+pow(*position->y, 2.0f))) * 180.0f / PI;
	else
		*current_point->azimuth = 180 - (asin(*position->y / sqrt(pow(*position->x,2.0f)+pow(*position->y, 2.0f))) * 180.0f / PI);
	//Just in case
	*current_point->polar = to_deg(*current_point->polar);
	*current_point->azimuth = to_deg(*current_point->azimuth);
}

