/*
 * model.h
 *
 *      Author: David Fernandez Lopez
 *      Debris Collector Core
 */

#include <vector>

#ifndef MODEL_H
#define MODEL_H

/*
Class point: represents any point in space using spherical coordinates.
*/
class Point{
	public:
		float *r, *polar, *azimuth;
		Point();
		Point(float radial_distance, float polar_angle, float azimuthal_angle);
		~Point();
};

/*
Class Orbit: represents an orbit described as a starting point and speed for
radial distance, polar angle and azimuthal angle. It contains the necesary 
methods to calculate the resulting point after a given time.
*/
class Orbit{
	public:
		Point *starting_point;
		float *v, *alpha, *beta;
		Orbit();
		Orbit(float r,float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed);
		~Orbit();
		Point move(float* time);
		void move(float* time, Point* p);
		void updateStartingPoint(float r, float polar, float azimuth);
};

/*
Class SpaceObject: represents an object in space, with it's orbit,
its current position, its mass and its starting time, when the object was added
to the model. Additionaly it informs if this object is a target or an
obstacle. 
*/
class SpaceObject{
	public:
		int* id;
		Point *current_point;
		Orbit *orbit;
		float *starting_time;
		SpaceObject(int new_id, float r, float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed, float time);
		~SpaceObject();
		void updateSpaceObject(float* time);
		bool exist(float time);
};

/*
Class TargetSO: represents a target space object. It inherits from SpaceObject.
 */
class TargetSO: public SpaceObject{
	public:
	int* object_mass;
	TargetSO(int id, float r, float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed, float time, int mass);
	~TargetSO();
};

/*
Class ObstacleSO: represents an obstacle space object. It inherits from SpaceObject.
 */
class ObstacleSO: public SpaceObject{
	public:
	ObstacleSO(int id, float r, float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed, float time);
	~ObstacleSO();
};

/*
Class ProjectableSpaceObject
*/
class ProjectableSpaceObject: public SpaceObject{
	Point* projected_point;
	float* projected_time;

	public:
	bool active;
	ProjectableSpaceObject(int new_id, float r, float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed);
	~ProjectableSpaceObject();
	Point* project(float time);
};

/*
Class ProjectableTargetSO
*/
class ProjectableTargetSO: public ProjectableSpaceObject{
	public:
	int* object_mass;
	ProjectableTargetSO(int new_id, float r, float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed, int mass);
	~ProjectableTargetSO();
};

/*
Class ProjectableObstacleSO
*/
class ProjectableObstacleSO: public ProjectableSpaceObject{
	public:
	ProjectableObstacleSO(int new_id, float r, float polar, float azimuth, float speed, float polar_speed, float azimuthal_speed);
	~ProjectableObstacleSO();
};

/*
Class Task
*/

class Task{
	public:
	int* target_id;
	float* time;
	Point* target_point;
	Task(int target, float dest_time, const std::vector<ProjectableSpaceObject*> &objects);
	~Task();
};

class EuclideanVector{
	public:
	float* x;
	float* y;
	float* z;

	EuclideanVector();
	EuclideanVector(float x0, float y0, float z0);
	~EuclideanVector();
	void copy_data(EuclideanVector* v);
};

class Robot{
	public:
		Point* current_point;
		EuclideanVector* movement;
		Robot();
		Robot(float r, float polar, float azimuth, float vx, float vy, float vz);
		~Robot();
		void move(float speed);
};

struct Beacon{
	int id;
	float distance;
};

/*
General methods
*/
float distance(Point* a, Point* b);
bool collision(Point* a, Point* b);
float to_deg(float angle);
float norm(EuclideanVector* v);
void quaternion_to_rot(float x, float y, float z, float w, std::vector<std::vector<float> >& rotation);
void apply_rotation(EuclideanVector* v1, std::vector<std::vector<float> >& rotation);

#endif
