/*
 * particle.cpp
 *
 *      Author: David Fernandez Lopez
 *      Debris Collector Core
 */

#include <cmath>
#include <vector>
#include <cstdlib>
#include "model.h"
#include "particle.h"

#define LIST_SIZE 3786
#define SCATTER_PROB 5
#define SPEED_MUT_PROB 10
#define PI 3.14159265f
#define MAX_ERROR 150000.0
#define MIN_PROBABILITY 0.0001f
#define BIG_ERROR 50000.0
#define BE_PROBABILITY 0.10f
#define SMALL_ERROR 10000.0
#define SE_PROBABILITY 0.50f


using namespace std;


/*******************************************/
/************ General use methods **********/
/*******************************************/

float probability(double error){
	if(error>MAX_ERROR){
		return MIN_PROBABILITY;
	}else if(error>BIG_ERROR){
		return ((BE_PROBABILITY - MIN_PROBABILITY)/(BIG_ERROR - MAX_ERROR))*(error - MAX_ERROR) + MIN_PROBABILITY;
	}else if(error>SMALL_ERROR){
		return ((SE_PROBABILITY - BE_PROBABILITY)/(SMALL_ERROR - BIG_ERROR))*(error - BIG_ERROR) + BE_PROBABILITY;
	}else{
		return ((1.0f - SE_PROBABILITY)/(0.0f - SMALL_ERROR))*(error - SMALL_ERROR) + SE_PROBABILITY;
	}
}

bool visible(Point* p1, Point* p2){
	float pol_diff = abs(*p1->polar - *p2->polar);
	float pol_diff_comp = 360.0f - pol_diff;
	float azi_diff = abs(*p1->azimuth - *p2->azimuth);
	float azi_diff_comp = 360.0f - azi_diff;

	return (pol_diff<=90.0f || pol_diff_comp <= 90.0f) && (azi_diff<=90.0f || azi_diff_comp<=90.0f);
}

float random_float(float a, float b) {
    float random = ((float) rand()) / (float) RAND_MAX;
    float diff = b - a;
    float r = random * diff;
    return a + r;
}

/*******************************************/
/****************** Classes ****************/
/*******************************************/

/**Class Particle methods**/
Particle::Particle(){
	current_point = new Point();
	mov = new EuclideanVector();
	fitness = new float;
	*fitness = 0.0f;
	error = new double;
	*error = 0.0;
	prev_error = new double;
	*prev_error = 1000000.0f;
}

Particle::Particle(float r, float polar, float azimuth, float vx, float vy, float vz){
	current_point = new Point(r, polar, azimuth);
	mov = new EuclideanVector(vx, vy, vz);
	float speed = norm(mov);
	*mov->x = *mov->x/speed;
	*mov->y = *mov->y/speed;
	*mov->z = *mov->z/speed;
	fitness = new float;
	*fitness = 0.0f;
	error = new double;
	*error = 0.0;
	prev_error = new double;
	*prev_error = 1000000.0f;
}

Particle::~Particle(){
	delete current_point;
	delete mov;
	delete fitness;
	delete error;
	delete prev_error;
	for(int i=0; i<distances.size(); i++){
		delete distances[i]->point;
		delete distances[i];
	}
	distances.clear();
	std::vector<Distance*>().swap(distances);
}

void Particle::loadDistance(Point* p, float dist){
	distances.push_back(new Distance);
	distances[distances.size()-1]->point = new Point(*p->r, *p->polar, *p->azimuth);
	distances[distances.size()-1]->expected_distance = dist;
}

void Particle::loadDistance(float r, float polar, float azimuth, float dist){
	distances.push_back(new Distance);
	distances[distances.size()-1]->point = new Point(r, polar, azimuth);
	distances[distances.size()-1]->expected_distance = dist;
}

void Particle::move_particle(float speed, vector<vector<float> > rotation){
	/*Translate to vector*/
	EuclideanVector* position = new EuclideanVector(*current_point->r * sin(*current_point->polar*PI/180.0f) * cos(*current_point->azimuth*PI/180.0f), *current_point->r * sin(*current_point->polar*PI/180.0f) * sin(*current_point->azimuth*PI/180.0f), *current_point->r * cos(*current_point->polar*PI/180.0f));

	/*Apply movement*/
	apply_rotation(mov, rotation);
	*position->x += speed*(*mov->x);
	*position->y += speed*(*mov->y);
	*position->z += speed*(*mov->z);

	/*Translate back*/
	*current_point->r = norm(position);
	*current_point->polar = acos(*position->z / *current_point->r) * 180.0f / PI;
	if(*position->x == 0.0f && *position->y == 0.0f)
			*current_point->azimuth = 0.0f;
	else if(*position->x >= 0.0f){
		*current_point->azimuth = to_deg((asin(*position->y / sqrt(pow(*position->x,2.0f)+pow(*position->y, 2.0f)))) * (180.0f / PI));
	}else{
		*current_point->azimuth = to_deg(360.0f - ((asin(*position->y / sqrt(pow(*position->x,2.0f)+pow(*position->y, 2.0f)))) * (180.0f / PI)));
	}

/*Compute distances*/
	for(unsigned int i=0; i< distances.size(); i++){
		if(visible(current_point, distances[i]->point)){
			distances[i]->real_distance = distance(current_point, distances[i]->point);
			*error += (double)abs(distances[i]->expected_distance - distances[i]->real_distance);
		}else{
			distances[i]->real_distance = 10000000.0f;
			*error += (double)10000000.0f;
		}
	}
	delete position;
}

void Particle::load(Particle* particle){
	int scatter = rand()%100;
	if(*particle->prev_error < *particle->error){
		*current_point->r = *particle->current_point->r + (scatter<=SCATTER_PROB?random_float(-100.0f,100.0f):0.0f);
		*current_point->polar = *particle->current_point->polar + (scatter<=SCATTER_PROB?random_float(-2.0f,2.0f):0.0f);
		*current_point->azimuth = *particle->current_point->azimuth + (scatter<=SCATTER_PROB?random_float(-2.0f,2.0f):0.0f);
		*mov->x = random_float(-2.0f,2.0f);
		*mov->y = random_float(-2.0f,2.0f);
		*mov->z = random_float(-2.0f,2.0f);
		float n = norm(mov);
		*mov->x = *mov->x / n;
		*mov->y = *mov->y / n;
		*mov->z = *mov->z / n;
	}else{
		*current_point->r = *particle->current_point->r;
		*current_point->polar = *particle->current_point->polar;
		*current_point->azimuth = *particle->current_point->azimuth;
		*mov->x = *particle->mov->x;
		*mov->y = *particle->mov->y;
		*mov->z = *particle->mov->z;
	}
	*prev_error = *particle->error;
}

/**Class ParticleList methods**/

ParticleList::ParticleList(bool random){
	if(!random){
		for(int i=0; i<LIST_SIZE; i++){
			particles.push_back(new Particle());
		}
	}else{
		for(float rad=100000.0f; rad<=200000.0f; rad+=50000.0f){
			for(float az=0.0f; az<360.0f; az+=10.0f){
				for(float po=10.0f; po<360.0f; po+=10.0f)
					particles.push_back(new Particle(rad,po,az,random_float(-2.0f,2.0f),random_float(-2.0f,2.0f),random_float(-2.0f,2.0f)));
			}
		}
		particles.push_back(new Particle(100000.0f,0.0f,0.0f,random_float(-2.0f,2.0f),random_float(-2.0f,2.0f),random_float(-2.0f,2.0f)));
		particles.push_back(new Particle(100000.0f,0.0f,180.0f,random_float(-2.0f,2.0f),random_float(-2.0f,2.0f),random_float(-2.0f,2.0f)));
		particles.push_back(new Particle(150000.0f,0.0f,0.0f,random_float(-2.0f,2.0f),random_float(-2.0f,2.0f),random_float(-2.0f,2.0f)));
		particles.push_back(new Particle(150000.0f,0.0f,180.0f,random_float(-2.0f,2.0f),random_float(-2.0f,2.0f),random_float(-2.0f,2.0f)));
		particles.push_back(new Particle(200000.0f,0.0f,0.0f,random_float(-2.0f,2.0f),random_float(-2.0f,2.0f),random_float(-2.0f,2.0f)));
		particles.push_back(new Particle(200000.0f,0.0f,180.0f,random_float(-2.0f,2.0f),random_float(-2.0f,2.0f),random_float(-2.0f,2.0f)));
	}
}

ParticleList::~ParticleList(){
	for(int i=0; i<particles.size(); i++)
		delete particles[i];
	particles.clear();
	std::vector<Particle*>().swap(particles);
}

double ParticleList::move(float speed, vector<vector<float> > rotation){
	float total_fitness = 0.0f;
	double total_error = 0.0;
	for(unsigned int i=0; i<particles.size(); i++){
		particles[i]->move_particle(speed, rotation);
		*particles[i]->fitness = probability(*particles[i]->error);
		total_fitness += *particles[i]->fitness;
		total_error += *particles[i]->error;
	}

	/*Normalize fitness*/
	*particles[0]->fitness = *particles[0]->fitness / total_fitness;
	for(unsigned int i=1; i<particles.size(); i++){
		*particles[i]->fitness = (*particles[i]->fitness / total_fitness) + *particles[i-1]->fitness;
	}
	return total_error/(double)particles.size();
}

int ParticleList::resample(){
	float chosen = random_float(0.0f,*particles[particles.size()-1]->fitness);
	for(unsigned int i=0; i<particles.size(); i++){
		if(chosen <= *particles[i]->fitness)
			return i;
	}
	return -1;
}

void ParticleList::computeMedium(Point* p, EuclideanVector* v){
	double r = 0.0, polar = 0.0, azimuth = 0.0, vx = 0.0, vy = 0.0, vz = 0.0;
	for(unsigned int i=0; i< particles.size(); i++){
		r += (double)*particles[i]->current_point->r;
		polar += (double)*particles[i]->current_point->polar;
		azimuth += (double)*particles[i]->current_point->azimuth;
		vx += (double)*particles[i]->mov->x;
		vy += (double)*particles[i]->mov->y;
		vz += (double)*particles[i]->mov->z;
	}

	*p->r = (float)(r / (double)particles.size());
	*p->polar = (float)(polar / (double)particles.size());
	*p->azimuth = (float)(azimuth / (double)particles.size());
	*v->x = (float)(vx / (double)particles.size());
	*v->y = (float)(vy / (double)particles.size());
	*v->z = (float)(vz / (double)particles.size());

	float speed = norm(v);
	*v->x = *v->x / speed;
	*v->y = *v->y / speed;
	*v->z = *v->z / speed;
}
