/*
 * planner.cpp
 *
 *      Author: David Fernandez Lopez
 *      Debris Collector Core
 */

#include <cstdlib>
#include <ctime>
#include <cmath>
#include <vector>
#include "model.h"
#include "genetics.h"
#include "ros/ros.h"
#include "debris_collector/SpaceObjectMsg.h"
#include "debris_collector/PlanTaskMsg.h"
#include "debris_collector/LocationMsg.h"
#include "std_srvs/Empty.h"

//Parameters
#define MAX_ITERATIONS 10000
#define MAX_DIFERENTIAL 0.005f
#define CONSECUTIVE_BESTS 10
#define MUTATION_PROBABILITY 5
#define MIGRATION_PROBABILITY 5
#define TIME_MUTATION_PROBABILITY 15
#define MAX_TIME_STEP 86400
#define MIN_TIME_STEP 60.0f

std::vector<ProjectableTargetSO*> targets;
Robot* robot = NULL;
bool genetic = false;
bool data;

/*Service callback*/
bool request_plan(std_srvs::Empty::Request &req, std_srvs::Empty::Response &res){
	genetic = true;
	return true;
}

//Mesage callback
void locationCallback(const debris_collector::LocationMsg& msg){
	*robot->current_point->r = msg.r;
	*robot->current_point->polar = msg.polar;
	*robot->current_point->azimuth = msg.azimuth;
	*robot->movement->x = msg.heading_x;
	*robot->movement->y = msg.heading_y;
	*robot->movement->z = msg.heading_z;
	data = true;
}

void objectsCallback(const debris_collector::SpaceObjectMsg& msg){
	//ROS_INFO_STREAM("Message recieved");
	if(msg.id%2!=0){
		for(unsigned int i=0; i<targets.size(); i++){
			if(*targets[i]->id == msg.id){
				//If the object is already in the targets vector some parameters should be updated
				targets[i]->orbit->updateStartingPoint(msg.r,msg.polar,msg.azimuth);
				targets[i]->active = true;
				return;
			}
		}
		//If the object has not yet appered in the targets vector it has to be inserted
		targets.push_back(new ProjectableTargetSO(msg.id,msg.r,msg.polar,msg.azimuth,msg.v,msg.alpha,msg.beta,msg.mass));
	}
}

/*General use methods*/
void geneticAlgorithm(ros::Publisher plan_pub){
	ROS_INFO_STREAM("Initiating genetic algorithm");
	//Create the initial random population
	Population* population = new Population(targets.size(), targets, robot->current_point);

	//Until convergence
	float best = 10000000000000.0f;
	int best_index = 0;
	int convergence = 0;
	int iterations = 0;
	while(iterations < MAX_ITERATIONS){
		//Calculate each individual cost
		best_index = population->compute_costs(targets, robot->current_point);

		if(std::abs(*population->best_cost-best)/std::max(*population->best_cost, best) <= MAX_DIFERENTIAL){
			if(convergence++ >= CONSECUTIVE_BESTS)
				break;
		}else{
			convergence = 0;
		 	if(*population->best_cost<best)
				best = *population->best_cost;
		}

		//Create new population
		Population* new_population = new Population(targets.size());
		for(unsigned int i=0; i<population->individuals.size(); i++){
			if(std::rand()%100<MIGRATION_PROBABILITY){
				//The new individual will be a randomly generated individual
				new_population->individuals[i]->migration(targets, robot->current_point);
			}else{
				int mother = population->chooseParent();
				int father = population->chooseParent();
				if(mother==-1) mother = best_index;
				if(father==-1) father = best_index;

				//The first element of the new individual is the first element in the parent wiht lower total cost
				if(*population->individuals[mother]->total_cost <= *population->individuals[father]->total_cost){
					new_population->individuals[i]->order[0] = population->individuals[mother]->order[0];
					new_population->individuals[i]->time[0] = population->individuals[mother]->time[0];
				}else{
					new_population->individuals[i]->order[0] = population->individuals[father]->order[0];
					new_population->individuals[i]->time[0] = population->individuals[father]->time[0];
				}
				//For the next element the one with lower cost between both parents will be chosen
				for(unsigned int j=1; j<new_population->individuals[i]->order.size(); j++){
					if(population->individuals[mother]->cost_to_next(new_population->individuals[i]->order[j-1]) <= population->individuals[father]->cost_to_next(new_population->individuals[i]->order[j-1])){
						//If this element is already in the order the other parent is chosen
						if(!new_population->individuals[i]->loadValue(population->individuals[mother]->next(new_population->individuals[i]->order[j-1]),population->individuals[mother]->next_time(new_population->individuals[i]->order[j-1]),j)){
							if(!new_population->individuals[i]->loadValue(population->individuals[father]->next(new_population->individuals[i]->order[j-1]),population->individuals[father]->next_time(new_population->individuals[i]->order[j-1]),j)){
								//If both parents are already present a random next value is selected
								new_population->individuals[i]->loadRandomValue(j);
							}
						}
					}else{
						//If this element is already in the order the other parent is chosen
						if(!new_population->individuals[i]->loadValue(population->individuals[father]->next(new_population->individuals[i]->order[j-1]),population->individuals[father]->next_time(new_population->individuals[i]->order[j-1]),j)){
							if(!new_population->individuals[i]->loadValue(population->individuals[mother]->next(new_population->individuals[i]->order[j-1]),population->individuals[mother]->next_time(new_population->individuals[i]->order[j-1]),j)){
								//If both parents are already present a random next value is selected
								new_population->individuals[i]->loadRandomValue(j);
							}
						}
					}
				}
				//Calculate order mutation
				if(std::rand()%100<=MUTATION_PROBABILITY){
					new_population->individuals[i]->mutation(std::rand()%new_population->individuals[i]->order.size(),std::rand()%new_population->individuals[i]->order.size());
				}
				//Calculate time mutation
				if(std::rand()%100<=TIME_MUTATION_PROBABILITY){
					//Load times
					new_population->individuals[i]->time[0] = loadTime(robot->current_point, targets[new_population->individuals[i]->order[0]], 0.0f);
					for(unsigned int j=1; j<new_population->individuals[i]->order.size(); j++){
						new_population->individuals[i]->time[j] = loadTime(targets[new_population->individuals[i]->order[j-1]]->project(new_population->individuals[i]->time[j-1]), targets[new_population->individuals[i]->order[j]], new_population->individuals[i]->time[j-1]);
					}
				}
			}

		}

		delete population;
		population = new_population;
		iterations++;
	}

	if(iterations == MAX_ITERATIONS)
		best_index = population->compute_costs(targets, robot->current_point);

	//Publish the selected plan
	for(unsigned int i=0; i<population->individuals[best_index]->order.size(); i++){
		ROS_INFO_STREAM(i << " target id " << *targets[population->individuals[best_index]->order[i]]->id << " at time " << population->individuals[best_index]->time[i]);
		debris_collector::PlanTaskMsg msg;
		msg.task_number = i;
		msg.target_id = *targets[population->individuals[best_index]->order[i]]->id;
		if(i==0)
			msg.destination_time = population->individuals[best_index]->time[i];
		else
			msg.destination_time = population->individuals[best_index]->time[i] - population->individuals[best_index]->time[i-1];
		msg.total_tasks = population->individuals[best_index]->order.size();
		plan_pub.publish(msg);
	}

	delete population;

	ROS_INFO_STREAM("Genetic algorithm finished");
}


int main(int argc, char **argv){
	//Loops per second, Hertz
	int time_ratio = 1;

	std::srand(time(0));
	robot = new Robot(0.0f,0.0f,0.0f,0.0f,0.0f,0.0f);
	data = false;

	//ROS init
	ros::init(argc, argv, "PlannerNode");
	ros::NodeHandle n;

	//ROS communication init
	ros::Subscriber sub = n.subscribe("Objects", 1000, objectsCallback);
	ros::Publisher plan_pub = n.advertise<debris_collector::PlanTaskMsg>("Plan", 1000);
	ros::Subscriber subLoc = n.subscribe("Location", 1000, locationCallback);
	ros::ServiceServer service = n.advertiseService("NewTaskList", request_plan);

	//ROS loop rate sets sleep time
	ros::Rate loop_rate(time_ratio);

	//The node loop, it will go on until the node is killed .
	//ros::ok() returns false if: a SIGINT is received (Ctrl-C), The node has been kicked off the network by another node with the same name,
	//ros::shutdown() has been called by another part of the application or all ros::NodeHandles have been destroyed.
	while(ros::ok()){
		if(data){
			//Targets vector cleanup, remove unactive nodes
			for(unsigned int i=0; i< targets.size(); i++){
				if(targets[i]->active){
					//No cleanup needed
					targets[i]->active = false;
				}else{
					//Cleanup needed, the projectable target is removed from the vector
					targets.erase(targets.begin()+i);
				}
			}

			//If there are more than a single target to catch, the current robot position has been calculated and 
			//a plan has been requested execute the genetic algorithm.
			if(genetic && targets.size()>1 && *robot->current_point->r!=0.0f){
				geneticAlgorithm(plan_pub);
				genetic = false;
			}
		}else{
			ROS_INFO_STREAM("No data recieved yet.");
		}

		//All pending messages are read and its callback methods called
		ros::spinOnce();
		loop_rate.sleep();
	}

	delete robot;
	for(int i=0; i<targets.size(); i++)
		delete targets[i];
	targets.clear();
	std::vector<ProjectableTargetSO*>().swap(targets);
}
