/*
 * world.cpp
 *
 *      Author: David Fernandez Lopez
 *      Debris Collector Core
 */

#include <vector>
#include <algorithm>
#include <cmath>
#include "ros/ros.h"
#include "debris_collector/SpaceObjectMsg.h"
#include "debris_collector/BeaconMsg.h"
#include "debris_collector/MovementMsg.h"
#include "debris_collector/GrabObjectSrv.h"
#include "model.h"

//Parameters
#define SENSOR_RANGE 150000.0f
#define GRAB_DISTANCE 500.0f

Robot* robot;
float speed = 0.5f;
std::vector<TargetSO*> targets;
std::vector<ObstacleSO*> obstacles;

/*Service callback*/
bool grab(debris_collector::GrabObjectSrv::Request &req, debris_collector::GrabObjectSrv::Response &res){
	for(unsigned int i=0; i<targets.size(); i++){
		if(*targets[i]->id == req.id){
			if(distance(targets[i]->current_point, robot->current_point)<GRAB_DISTANCE){
				res.result = 0;
				targets.erase(targets.begin()+i);
				return true;
			}
		}
	}
	res.result = 1;
	return true;
}


/*Topic callback*/
void moveCallback(const debris_collector::MovementMsg& msg){
	std::vector<std::vector<float> > rotation;
	rotation.resize(3);
	rotation[0].resize(3);
	rotation[1].resize(3);
	rotation[2].resize(3);
	quaternion_to_rot(msg.x, msg.y, msg.z, msg.w, rotation);
	apply_rotation(robot->movement, rotation);
	speed = msg.speed;
}

/*General use methods*/
bool compareBeacons(Point* b1, Point* b2){
	return distance(b1, robot->current_point) < distance(b2, robot->current_point);
}

void translate(debris_collector::SpaceObjectMsg* msg, TargetSO* target){
	msg->id = *target->id;
	msg->r = *target->current_point->r;
	msg->polar = *target->current_point->polar;
	msg->azimuth = *target->current_point->azimuth;
	msg->v = *target->orbit->v;
	msg->alpha = *target->orbit->alpha;
	msg->beta = *target->orbit->beta;
	msg->mass = *target->object_mass;
}

void translate(debris_collector::SpaceObjectMsg* msg, ObstacleSO* obstacle){
	msg->id = *obstacle->id;
	msg->r = *obstacle->current_point->r;
	msg->polar = *obstacle->current_point->polar;
	msg->azimuth = *obstacle->current_point->azimuth;
	msg->v = *obstacle->orbit->v;
	msg->alpha = *obstacle->orbit->alpha;
	msg->beta = *obstacle->orbit->beta;
	msg->mass = 0;
}

int main(int argc, char **argv){
	//Loops per second, Hertz
	int time_ratio = 1;

	//Vector of Targets to catch, target id should always be odd and different from 0
	targets.push_back(new TargetSO(1, 145000.0f, 20.0f, 25.0f, 0.0f, 0.004f,-0.005f,0.0f,1000));
	targets.push_back(new TargetSO(3, 155000.0f, 90.0f, 330.0f, 0.0f, -0.003f,0.004f,0.0f,1000));

	//Vector of Obstacles to avoid, obstacle id should always be even and different from 0
	obstacles.push_back(new ObstacleSO(2, 145000.0f, 0.0f, 90.0f, 0.0f, 0.003f,-0.01f,0.0f));
	obstacles.push_back(new ObstacleSO(4, 155000.0f, 90.0f, 0.0f, 0.0f, -0.005f,0.001f,0.0f));

	//Current Robot position
	robot = new Robot(150000.0f,45.0f,0.0f,1.0f/std::sqrt(3.0f),1.0f/std::sqrt(3.0f),1.0f/std::sqrt(3.0f));

	//Beacons
	std::vector<Point*> beacons;
	beacons.push_back(new Point(50000.0f,0.0f,0.0f));
	beacons.push_back(new Point(50000.0f,180.0f,0.0f));
	beacons.push_back(new Point(50000.0f,90.0f,0.0f));
	beacons.push_back(new Point(50000.0f,90.0f,90.0f));
	beacons.push_back(new Point(50000.0f,90.0f,180.0f));
	beacons.push_back(new Point(50000.0f,90.0f,270.0f));

	//Current time
	float timer = 0.0f;

	//ROS init
	ros::init(argc, argv, "WorldNode");
	ros::NodeHandle n;

	//ROS communication init
	ros::Publisher objects_pub = n.advertise<debris_collector::SpaceObjectMsg>("Objects", 1000);
	ros::Publisher beacons_pub = n.advertise<debris_collector::BeaconMsg>("Beacon", 1000);
	ros::Subscriber subMov = n.subscribe("Move", 1000, moveCallback);
	ros::ServiceServer service = n.advertiseService("Grab", grab);

	//ROS loop rate sets sleep time
	ros::Rate loop_rate(time_ratio);

	//The node loop, it will go on until the node is killed .
	//ros::ok() returns false if: a SIGINT is received (Ctrl-C), The node has been kicked off the network by another node with the same name,
	//ros::shutdown() has been called by another part of the application or all ros::NodeHandles have been destroyed.
	while(ros::ok()){
		//The robot moves in the direction of its orientation at the provided speed
		robot->move(speed);
		ROS_INFO_STREAM("At Second: "<< timer <<"s " << "Robot "<<*robot->current_point->r << "," << *robot->current_point->polar << "," << *robot->current_point->azimuth << "-" << *robot->movement->x << ","<< *robot->movement->y << ","<< *robot->movement->z);

		//Beacons are sorted to choose the nearest 3 beacons, which will be sent to the locator node
		std::sort(beacons.begin(), beacons.end(), compareBeacons);
		for(int i=0; i<3; i++){
			debris_collector::BeaconMsg msg;
			msg.beacon_r = *beacons[i]->r;
			msg.beacon_polar = *beacons[i]->polar;
			msg.beacon_azimuth = *beacons[i]->azimuth;
			msg.distance = distance(beacons[i], robot->current_point);
			beacons_pub.publish(msg);
		}

		//All target positions are updated
		for(int i = 0; i<targets.size(); i++){
			if(targets[i]->exist(timer)){
				targets[i]->updateSpaceObject(&timer);
				if(distance(robot->current_point, targets[i]->current_point)<=SENSOR_RANGE){
					//If the target is in sensor range it is published in the objects topic
					ROS_INFO_STREAM("Publishing target "<< *targets[i]->id << " ("<<*targets[i]->current_point->r<<","<<*targets[i]->current_point->polar<<","<<*targets[i]->current_point->azimuth<<")");
					debris_collector::SpaceObjectMsg msg;
					translate(&msg, targets[i]);
					objects_pub.publish(msg);
				}
			}
		}

		//All obstacle positions are updated
		for(int i = 0; i<obstacles.size(); i++){
			if(obstacles[i]->exist(timer)){
				obstacles[i]->updateSpaceObject(&timer);
				//If the robot comes to close to an obstacle a collision is reported
				ROS_ERROR_STREAM_COND(collision(robot->current_point, obstacles[i]->current_point),"Collision with obstacle "<<*obstacles[i]->id<<" at point ("<<*robot->current_point->r<<","<<*robot->current_point->polar<<","<<*robot->current_point->azimuth<<") second "<<timer<<"");
				if(distance(robot->current_point, obstacles[i]->current_point)<=SENSOR_RANGE){
					//If the obstacle is in sensor range it is published in the objects topic
					ROS_INFO_STREAM("Publishing obstacle "<< *obstacles[i]->id << " ("<<*obstacles[i]->current_point->r<<","<<*obstacles[i]->current_point->polar<<","<<*obstacles[i]->current_point->azimuth<<")");
					debris_collector::SpaceObjectMsg msg;
					translate(&msg, obstacles[i]);
					objects_pub.publish(msg);
				}
			}
		}

		//Timer is updated
		timer += 1.0f;
		//All pending messages are read and its callback methods called
		ros::spinOnce();
		loop_rate.sleep();
	}

	delete robot;
	for(int i=0; i<targets.size(); i++)
		delete targets[i];
	targets.clear();
	std::vector<TargetSO*>().swap(targets);
	for(int i=0; i<obstacles.size(); i++)
		delete obstacles[i];
	obstacles.clear();
	std::vector<ObstacleSO*>().swap(obstacles);
	for(int i=0; i<beacons.size(); i++)
		delete beacons[i];
	beacons.clear();
	std::vector<Point*>().swap(beacons);

	return 0;
}
